/*     DRAWxtl                                                           */
/*                                                                       */
/* Routine to read crystallographic information such as unit cell, space */
/* group, atom positions, and thermal ellipsoid parameters to generate   */
/* POLYHEDRON, SPHERE, ELLIPSOID and BOND geometry POV and VRML scenes.  */
/*                                                                       */
/* Part 2                                                                */
/*                                                                       */
/* Coded by:   L.W. Finger                                               */
/*            Geophysical Laboratory                                     */
/*            Carnegie Institution of Washington                         */
/*            Washington DC 20015                                        */
/*            finger@gl.ciw.edu                                          */
/*                                                                       */
/* Original Version: April 92                                            */
/*            Revised January 1993 to add bonds.                         */
/*            Converted to AVS Module February 1993.                     */
/*            Thermal Ellipsoids added April 1996                        */
/*            Thermal ellipsoids for POV and VRML finished Nov. 1996     */
/*                                                                       */
/*                                                                       */
/* Converted to standalone POV scene file generator Oct. 1994/March 1995 */
/* and VRML generator added September 1995/March 1996                    */
/* by M. Kroeker, Organic Chemistry, Technische Hochschule Darmstadt,    */
/* Petersenstr. 22, D-64287 Darmstadt, Germany                           */
/* martin@oc2.oc.chemie.th-darmstadt.de                                  */
/*                                                                       */
/* Two versions merged  July 1995                                        */
/* ===================================================================== */

/*
 This file contains routines: 
    generate_ellipsoids, generate_planes, generate_poly,
    import_fdat,import_gsas, import_shelx, Unique_Atom
*/

#include "drawxtl.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

extern float dist();

/* global variables for lattice constants and symmetry operators */

extern char spg[3][4];                 /* space group code */
extern int ss[24][3][3];               /* rotation operators for symmetry */
extern float lat_pos[4][3];            /* coordinates of lattice points */
extern float ts[24][3];                /* translation operators */
extern int sym_op_no[192];             /* saved symmetry operation number */
extern float cell_xyz[192][3];         /* storage for generated equiv. positions */
extern int sys;                        /* crystal system indicator */
extern int nbr;                        /* Bravais lattice indicator */
extern int ns;                         /* center of symmetry indicator */
extern int ng;                         /* number of rotation operations to specify group */
extern int nlat;                       /* number of lattice positions */
extern float lat_con[6];               /* lattice constants */
extern float ginv[3][3];               /* direct space metric tensor */
extern float b_inv_mat[3][3];          /* inverse metric tensor */
extern float b_mat[3][3];              /* metric tensor */
extern float rec_lat_con[3];           /* reciprocal lengths */

/* global variables for file names and file variables */

extern char  savedpath[100];           /* path at start of program */
extern char  File_Name[100];           /* Modified Input File Name */
extern FILE  *fpin,*flout;
extern FILE  *fpoutv;
extern FILE  *fpoutp;
extern char  *Str_File_Inp;            /* Input File Name */
extern char  *Str_File;
extern char  *output;

/* global variables for atom lists */

extern int   natom;                    /* number of different types of atoms */
extern int   ncell;                    /* number of atoms in asymmetric unit */
extern int   atom_no[MAX_VERTS];       /* master list atom and symmetry number */
extern float xypos[3*MAX_VERTS];       /* master list fractional coordinates */
extern float *o_vert;                  /* crystal coordinates of saved vertices */
extern float *s_vert;                  /* saved coordinates of vertices */
extern int   *vert_sym_no;             /* saved symmetry operator number */
extern int   *poly_list;               /* storage for polygon corner pointers */
extern int   nvert;                    /* number of vertices in list */
extern int   NvertM;                    /* number of vertices in master atom list */
extern float atom_xyz[MAX_ATOM][3];    /* atomic coordinates */
extern char  atom_l[MAX_ATOM][4];      /* atom labels */
extern int   atom_n[MAX_ATOM];         /* input atom number */
extern int   sv_atom_n[MAX_ATOM];      /* saved input atom number */

/* global variables for spheres and ellipsoids */

extern int   nsphere;                  /* number of different types of spheres */
extern int   n_ellips;                 /* no. of aniso. atoms in asymmetric unit */
extern int   do_ellipsoids;            /* non-zero if ellipsoids to be drawn */
extern int   El_Cutout;                /* True if ellipsoids to have an octant removed */
extern float Sphere_Mult;              /* Multiplier for Sphere Size */
extern float SpMult;                   /* initialization for Sphere_Mult */
extern float Ellipsoid_Prob;           /* Fraction of electron dens enclosed by ellipsoid */
extern float Ellipsoid_Scale;          /* Scaling needed to get probability */
extern float Phong_Value;              /* reflectivity value for Phong highlighting */
extern float Phong_Size;               /* Phong highlighting size */
extern char  sphere_l[MAX_SPHERE][4];  /* sphere label */
extern float sphere_size[MAX_SPHERE];  /* sphere size */
extern char  sphere_col_v[MAX_SPHERE][30];/* sphere colors */
extern char  sphere_col_p[MAX_SPHERE][30];/* sphere colors */
extern char  ellips_l[MAX_EL][4];      /* label for anisotropic atom */
extern int   ellips_n[MAX_EL];         /* atom number */
extern int   ell_type[MAX_EL];         /* form of ellipsoid input */
extern float ellips[MAX_EL][6];        /* thermal ellipsoid coefficients */
extern float ellips_RMS[MAX_EL][3];    /* RMS amplitudes */
extern float ellips_EV[MAX_EL][3][3];  /* Eigenvectors */
extern char  ellips_col_v[MAX_EL][30]; /* ellipsoid colors for VRML */
extern char  ellips_col_p[MAX_EL][30]; /* ellipsoid colors for POV */
extern char  Cutout_color[30];         /* color of cutout section */

/* global variables for polyhedra, planes and bonds */

extern int   npoly;                    /* number of different types of polyhedra */
extern int   nplane;                   /* number of different types of planes */
extern int   nbond;                    /* number of different types of bonds */
extern int   draw_list;                /* number of items in polygon draw list */
extern int   *vertex_list;             /* storage for vertices about a cation */
extern int   numb_list;                /* number in polygon list */
extern int   domolcomp;                /* non-zero if molecule completion requested */
extern float polyfudge;                /* should be close to zero to allow only regular polygonal faces */
extern float Bond_Mult;                /* Multiplier for Bond Size */
extern float BndMult;                  /* Bond multiplier initialization */
extern float mol_d;                    /* intramolecular distance for molecular and polyhedral completion */
extern char  poly_l[MAX_POLY][4];      /* polyhedral center label */
extern float poly_size[MAX_POLY];      /* size of polyhedron */
extern char  poly_col_v[MAX_POLY][30]; /*polyhedron colors for VRML */
extern char  poly_col_p[MAX_POLY][30]; /*polyhedron colors for POV */
extern char  plane_col_v[MAX_PLANE][30];/*plane colors for VRML */
extern char  plane_col_p[MAX_PLANE][30];/*plane colors for POV */
extern char  plane_l[MAX_PLANE][4];    /* plane center label */
extern float plane_size[MAX_PLANE];    /* size of plane coordination */
extern char  bond_l1[MAX_BOND][4];     /* label for one end of bond */
extern char  bond_l2[MAX_BOND][4];     /* label for other end */
extern float bond_size[MAX_BOND];      /* bond diameter */
extern float bond_min[MAX_BOND];       /* minimum bond distance */
extern float bond_max[MAX_BOND];       /* maximum bond distance */
extern char  col_bond_v[MAX_BOND][30]; /* color and properties of bonds for VRML */
extern char  col_bond_p[MAX_BOND][30]; /* color and properties of bonds for POV */
 
/* global variables that control size and orientation of output object */

extern float X_Boxlim;                 /* X box limit read from command line */
extern float Y_Boxlim;                 /* Y box limit read from command line */
extern float Z_Boxlim;                 /* Z box limit read from command line */
extern float X_Origin;                 /* X origin read from command line */
extern float Y_Origin;                 /* Y origin read from command line */
extern float Z_Origin;                 /* Z origin read from command line */
extern float boxlim[3];                /* half limits of plotting box */
extern float origin[3];                /* position of plotting origin */
extern float G_Rot[3][3];              /* Grand rotation matrix - transforms from Cartesian to picture */
extern float lookat_v1[3],lookat_v2[3];/* "lookat" vectors " */
extern float POV_Max[3];               /* Maximum limits of POV after rotation */
extern float POV_Min[3];               /* Minimum limits of POV after rotation */
extern int   boxflag;                  /* Non-zero if 'bounds' command given */
extern int   packflag;                 /* Non-zero if 'pack' command given */
extern int   docell;                   /* non-zero if unit-cell edges to be drawn */
extern int   Display_axes;             /* non-zero if axial triple to be drawn */
extern int   Color_Warning;            /* Warning flag for non-standard color */
extern float Magnification;            /* Magnification factor for image */
extern float Scale;                    /* Scale of diagram */
extern float offset[3];                /* offset for vector triple */
extern float Text_Size;                /* Size to make axis label text */
extern int   Options;                  /* Place to save command-line options flags */
extern char  col_cell_v[30];           /* color and properties of framebars */
extern char  col_edge_v[30];           /* color and properties of polyhedron edges */
extern char  col_cell_p[30];           /* color and properties of framebars */
extern char  col_edge_p[30];           /* color and properties of polyhedron edges */
extern float rad_cell;                 /* radius of unit cell framebars */
extern int   edges;                    /* draw thin black lines around the edges of polyhedra*/
extern float rad_edge;                 /* radius of cylinders on polyhedron edges */
extern float xrot,yrot,zrot;           /* view rotation angles */
extern float cryst_lim[6];             /* limits of view box in cell coords */
extern int   Unit_Cell;                /* non-zero if unit cell to be drawn */
extern int   no_comment;               /* non-zero to inhibit comment lines in VRML */
extern int   M_cameras;                /* non-zero to inhibit multiple cameras */

#define GEOMedit_list char
#define GEOMobj int

void Blank_Strip();
void build_poly_list();
float determinant();
void expand_atom();
void find_all_in_box();
void generate_ellipsoids();
void generate_planes();
void generate_poly();
void generate_spheres();
void import_cif();
void import_gsas();
void import_shelx();
void Output_Spheres();
void plot_vrml_poly();
void symop();
void Token_Strip();
int  toupper();
int  Unique_Atom();

/* ************************************************************** */
/* ************************************************************** */

float determinant(rot)

float rot[3][3];

{
  float temp;
  temp = rot[0][0] * (rot[1][1] * rot[2][2] - rot[1][2] * rot[2][1])
     - rot[1][0] * (rot[0][1] * rot[2][2] - rot[0][2] * rot[2][1])
     + rot[2][0] * (rot[0][1] * rot[1][2] - rot[0][2] * rot[1][1]);
  return temp;
}

/* ************************************************************** */
/* ************************************************************** */

void generate_ellipsoids(output)

GEOMedit_list *output;

/* routine to generate ellipsoid objects and add then to edit list */

{
  extern int natom;                 /* number of different types of atoms */
  extern int ss[24][3][3];          /* rotation operators for symmetry */
  float ev[3];                      /* scaled eigenvalue */
  float rot[3][3];                  /* rotation matrix for triangles */
  float rot2[3][3];                 /* matrices for checking visibility of cutout */
  int i,j,k,l,m,n,o,V1[3];          /* loop counters */
  int n_ellipsoids;
  float EV[3][3];                   /* array for eigenvectors */
  float det;
  float alpha1,alpha2,alpha3,sinal2; /* terms needed to calculate rotations */
  float d1,d2,d3,alpha,sinalp,radius;

  n_ellipsoids = 0;
  radius = 0.0005 * Scale;
  fprintf(flout,"\nGenerated %6.1f percent probability ellipsoids for:\n\n"
      ,100*Ellipsoid_Prob);
  for (i=1;i<n_ellips;++i) {    /* loop through ellipsoids */
    for (j=0;j<3;j++) {
      ev[j] = ellips_RMS[i][j] * Ellipsoid_Scale;
      for(k=0;k<3;k++)
        EV[j][k] = ellips_EV[i][j][k];
    }
    for (j=0;j<natom;++j) {  /* loop through atoms */
      nvert = 0;              /* clear the vertex list */
      if( atom_n[j] == -10000 - i) /* this negative means ellipsoid */
            find_all_in_box(j); /* generate all locations for this atom */
        if (nvert > 0)  {    /* if some present */
          if (do_ellipsoids == 0) { /* Biso sphere */
/* allocate storage for radii of spheres */
            float *radii;
	    radii = (float *) malloc ((unsigned)(nvert*sizeof(float)));
	    for (k=0;k<nvert;++k) /* copy radii info to storage*/
	      radii[k] = ellips_RMS[i][0] * Ellipsoid_Scale;
            Output_Spheres(radii,output,i);
            free(radii);
          } else {
            for (k=0;k<nvert;++k){ /* loop through the atom positions */
              o = vert_sym_no[k];
              fprintf(flout,"  %c%c%c%c at %8.5f %8.5f %8.5f \n"
                 ,atom_l[j][0],atom_l[j][1],atom_l[j][2],atom_l[j][3]
                 ,o_vert[3*k],o_vert[3*k+1],o_vert[3*k+2]);
              if (no_comment == 0) fprintf(fpoutv,"# Ellipsoid around %c%c%c%c at %8.5f %8.5f %8.5f \n"
                 ,atom_l[j][0],atom_l[j][1],atom_l[j][2],atom_l[j][3],o_vert[3*k],o_vert[3*k+1]
                 ,o_vert[3*k+2]);
              fprintf(fpoutp,"/* Ellipsoid around %c%c%c%c at %8.5f %8.5f %8.5f */\n"
                 ,atom_l[j][0],atom_l[j][1],atom_l[j][2],atom_l[j][3],o_vert[3*k],o_vert[3*k+1]
                 ,o_vert[3*k+2]);
              for (l=0;l<3;++l) {
                for (m=0;m<3;++m) {
                  rot[l][m] = 0.0;
                  for (n=0;n<3;++n) {
                    rot[l][m] += EV[l][n] * ss[o][n][m];
                  }  /* now have combined rotation matrix for ellipsoid */
                }    /* for on m */
              }      /* for on l */

/* make sure 'rot' is right handed - ss could be left-handed */

             det = determinant(rot);
             if (det < 0 )
                for (m=0;m<3;m++)
                  rot[0][m] *= -1.0;
/* get rid of very small components */
              for (l=0;l<3;++l)
                for (m=0;m<3;++m)
                  if (fabs(rot[l][m]) < 1.0e-5) rot[l][m] = 0.0;
              for (l=0;l<3;l++) {   /* perform the POV view rotation */
                for (m=0;m<3;m++) {
                  rot2[l][m] = 0.0;
                  for (n=0;n<3;n++)
                    rot2[l][m] += G_Rot[n][l] * rot[n][m];
                }
              }
              if (El_Cutout == 1) { 
                if (rot2[2][2] < 0.0) {  /* make cos(alpha2) >= 0 */
                  for (l=0;l<3;l++) {
                    rot2[2][l] *= -1.0;
                    rot2[1][l] *= -1.0;
                  }
                }
                if (rot2[0][2] > 0.0) {  /* make cos(alpha1) >= 0 */
                  for (l=0;l<3;l++) {
                    rot2[0][l] *= -1.0;
                    rot2[1][l] *= -1.0;
                  }
                }
                if (rot2[2][0] < 0.0) {  /* make cos(alpha3) >= 0 */
                  for (l=0;l<3;l++) {
                    rot2[l][0] *= -1.0;
                    rot2[l][1] *= -1.0;
                  }
                }
                if (rot2[1][2] < 0.0) {  /* make alpha1 < 0 */
                  for (l=0;l<3;l++) {
                    rot2[1][l] *= -1.0;
                    rot2[l][1] *= -1.0;
                  }
                }
                for (m=0;m<3;m++)        /* select which octant to dispplay */
                  if (rot2[2][m] < 0.0)
                    V1[m] = -1;
                  else
                    V1[m] = 1;
              }
              for (l=0;l<3;l++) {   /* perform the POV view rotation */
                for (m=0;m<3;m++) {
                  rot[l][m] = 0.0;
                  for (n=0;n<3;n++)
                    rot[l][m] += G_Rot[l][n] * rot2[n][m];
                }
              }
fprintf(fpoutp,"/* rot2:%8.5f %8.5f %8.5f */\n",rot2[0][0],rot2[0][1],rot2[0][2]);
fprintf(fpoutp,"/*      %8.5f %8.5f %8.5f */\n",rot2[1][0],rot2[1][1],rot2[1][2]);
fprintf(fpoutp,"/*      %8.5f %8.5f %8.5f */\n",rot2[2][0],rot2[2][1],rot2[2][2]);
fprintf(fpoutp,"/* rot: %8.5f %8.5f %8.5f */\n",rot[0][0],rot[0][1],rot[0][2]);
fprintf(fpoutp,"/*      %8.5f %8.5f %8.5f */\n",rot[1][0],rot[1][1],rot[1][2]);
fprintf(fpoutp,"/*      %8.5f %8.5f %8.5f */\n",rot[2][0],rot[2][1],rot[2][2]);
/* calculate rotation angles and direction cosines */
              alpha2 = -acos(rot[2][2]);
              sinal2 = sin(alpha2);
              if( fabs(sinal2) > 0.00001) {    
                alpha1 = atan2(rot[1][2]/sinal2,rot[0][2]/sinal2);
                alpha3 = atan2(rot[2][1]/sinal2,-rot[2][0]/sinal2);
              } else {
                alpha3 = atan2(rot[0][1],rot[0][0]);
                alpha1 = 0.0;
              }
              if (alpha1 < -90.0) alpha1 += 180.0;
              if (alpha1 > 90.0) alpha1 -= 180.0;
/* output this ellipsoid to the POV file */
              fprintf(fpoutp," object{\n");
              fprintf(fpoutp,"union{\n");
              fprintf(fpoutp," object{\n");
              if (El_Cutout)fprintf(fpoutp,"   difference {\n");
              fprintf(fpoutp,"   sphere{< 0,0,0 > 1}\n");
              if (El_Cutout) {
                fprintf(fpoutp,"   box{<0,0,0>,<%d,%d,%d>}\n",V1[0],V1[1],V1[2]);
              }
              fprintf(fpoutp,"   texture{pigment{color %s  }}\n",ellips_col_p[i]);
              fprintf(fpoutp,"   finish{phong %5.2f phong_size %5.2f reflection 0.1}\n"
                ,Phong_Value,Phong_Size);
              fprintf(fpoutp,"   scale<%8.5f,%8.5f,%8.5f>\n",ev[0],ev[1],ev[2]);
              if (El_Cutout)fprintf(fpoutp,"  }\n");
              fprintf(fpoutp," }\n");
              fprintf(fpoutp," object{\n");
              fprintf(fpoutp,"  sphere{< 0,0,0 > 1\n"); /* first principal ellipse */
              fprintf(fpoutp,"   texture{pigment{color %s}}\n",Cutout_color);
              fprintf(fpoutp,"   scale<%8.5f,%8.5f,%8.5f>\n  }\n }\n",ev[0]+.01,ev[1]+.01,radius);
              fprintf(fpoutp," object{\n  sphere{< 0,0,0 > 1\n"); /* second principal ellipse */
              fprintf(fpoutp,"   texture{pigment{color %s}}\n",Cutout_color);
              fprintf(fpoutp,"   scale<%8.5f,%8.5f,%8.5f>\n  }\n }\n",ev[0]+.01,radius,ev[2]+.01);
              fprintf(fpoutp," object{\n  sphere{< 0,0,0 > 1\n"); /* third principal ellipse */
              fprintf(fpoutp,"   texture{pigment{color %s}}\n",Cutout_color);
              fprintf(fpoutp,"   scale<%8.5f,%8.5f,%8.5f>\n  }\n }\n}\n",radius,ev[1]+.01,ev[2]+.01);
              if(fabs(sinal2) > 0.00001) {
                fprintf(fpoutp,"  rotate <0,0,%8.5f>\n",alpha1*RAD);
                fprintf(fpoutp,"  rotate <0,%8.5f,0>\n",alpha2*RAD);
                fprintf(fpoutp,"  rotate <0,0,%8.5f>\n",alpha3*RAD);
              } else {
                fprintf(fpoutp,"  rotate <0,0,%8.5f>\n",alpha3*RAD);
              }
              fprintf(fpoutp,"  translate <%8.5f, %8.5f, %8.5f>\n }\n"
                  ,s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
/* adjust 'rot' so that the trace is > 0 */
            if (rot[0][0] + rot[1][1] + rot[2][2] < 0)
              for (l=1;l<3;++l)
                for (m=0;m<3;++m)
                  rot[l][m] *= -1.0;
            if (rot[0][0] + rot[1][1] + rot[2][2] < 0)
              for (l=0;l<2;l++)
                for (m=0;m<3;m++)
                  rot[l][m] *= -1.0;

/* Output the ellipsoid to the VRML file */

              fprintf(fpoutv," Separator {\n");
              fprintf(fpoutv,"  Transform {\n");
              d1 = 0.5 * (rot[2][1] - rot[1][2]);
              d2 = 0.5 * (rot[0][2] - rot[2][0]);
              d3 = 0.5 * (rot[1][0] - rot[0][1]);
/* d's contain direction cosines * sin(alpha) */
              alpha = 1.0 - (d1 * d1 + d2 * d2 + d3 * d3);
              if (alpha < 0 ) alpha = 0.0;
              alpha = acos(sqrt(alpha));
              sinalp = sin(alpha);
              if (fabs(sinalp) > 0.0001) {
                float temp;
                d1 = d1/sinalp;
                d2 = d2/sinalp;
                d3 = d3/sinalp;

/* Make sure alpha is in correct quadrant */

                temp = cos(alpha) + (1.0 - cos(alpha)) * d1 * d1;
                if (fabs(rot[0][0] - temp) > 0.0001) alpha = 3.141597 - alpha;
                if (fabs(d1) > 0.0001 ) {
                  temp = (1.0 - cos(alpha)) * d2 * d3 - d1 * sin(alpha);
                  if (fabs(rot[1][2] - temp) > 0.0001) alpha = 6.283194 - alpha;
                } else {
                  if (fabs(d2) > 0.0001 ) {
                    temp = (1.0 - cos(alpha)) * d1 * d3 - d2 * sin(alpha);
                    if (fabs(rot[2][0] - temp) > 0.0001) alpha = 6.283194 - alpha;
                  } else {
                    if (fabs(d3) > 0.0001 ) {
                      temp = (1.0 - cos(alpha)) * d1 * d2 - d3 * sin(alpha);
                      if (fabs(rot[0][1] - temp) > 0.0001) alpha = 6.283194 - alpha;
                    }
                  }
                }
                fprintf(fpoutv,"   rotation %6.3f %6.3f %6.3f %7.3f\n",d1,d2,d3,-alpha);
              }
              fprintf(fpoutv,"   translation %8.5f %8.5f %8.5f\n",s_vert[3*k],s_vert[3*k+1],s_vert[3*k+2]);
              fprintf(fpoutv,"   scaleFactor %8.5f %8.5f %8.5f\n",ev[0],ev[1],ev[2]);
              fprintf(fpoutv,"  }\n");
              fprintf(fpoutv,"  Material {\n");
              fprintf(fpoutv,"   shininess 0.3\n");
              fprintf(fpoutv,"   diffuseColor  %s\n",ellips_col_v[i]);
              fprintf(fpoutv,"  }\n");
              fprintf(fpoutv,"  Sphere {\n");
              fprintf(fpoutv,"   radius     1\n  }\n");
              fprintf(fpoutv," }\n");
              n_ellipsoids++;
          }   /* for on k */
        }     /* if nvert > 0 */
      }    /* if ellipsoid */
    }	
  }
  printf("%4d ellipsoids\n",n_ellipsoids);
  fprintf(flout,"%4d ellipsoids generated\n",n_ellipsoids);
}  /* end of generate_ellipsoids */

/* ************************************************************** */
/* ************************************************************** */

void generate_planes(output)

GEOMedit_list *output;

/* procedure to generate corners of planes */

{
    extern int natom;                      /* number of different types of atoms */
    int i,j,k;                  /* loop counters, etc */
    int nvert_start;            /* starting number for cations */
    int planeno;                /* number of current polyhedron */
    int Plane_Count;            /* number of planes output */
    int ii,jj,ll,l;
    int *no,*ns,radius;
    float xm,ym,zm,vx,vy,vz,bm,vxs,vys,vzs,bms,phi,phi0,cosphi;

    radius = rad_edge * 0.01 * Scale;
    if (nplane == 1) return;     /* exit NOW if no work */
    Plane_Count = 0;
    for (i=0;i<3;++i) boxlim[i] += 3.0; /* expand box for cation generation */
    nvert = 0;                 /* make list empty */
    for (i=0;i<natom;++i)       /* loop through atoms */
        if (atom_n[i] < 1 ) find_all_in_box(i); /* generate all cations */
    for (i=0;i<3;++i) boxlim[i] -= 3.0; /*shrink box to original size */
    if (nvert == 0) return;     /* if no anions, no polyhedra */
    nvert_start = nvert;        /* save start of cation list */
    for (i=0;i<natom;++i) {     /* start loop for cations */
	nvert = nvert_start;    /* get rid of previous cations */
	if (atom_n[i] > 10000) {    /* greater than zero for polyhedral centers */
          planeno = atom_n[i]-10000;  /* save polyhedron number */
          find_all_in_box(i);  /* add all polyhedral centers in box */
          if(nvert_start < nvert) { /* check for some found */
  	    draw_list = 0;  /* initialize the draw list */
	    for (j=nvert_start;j<nvert;++j) { /* loop through cations */
	      numb_list = 0;      /* list clear */
	      for (k=0;k<nvert_start;++k) { /* loop through anions */
		if (dist(j,k) <= plane_size[planeno] && dist(j,k) > 0.01)  /* add vertex k to poly list */
	          vertex_list[numb_list++] = k;
	      }
	      if (numb_list > 2) {         /* only do if at least a triangle*/
                fprintf(flout,"Polygon with %d vertices around %c%c%c%c at %8.5f %8.5f %8.5f\n"
                  ,numb_list,atom_l[i][0],atom_l[i][1],atom_l[i][2],atom_l[i][3]
                  ,s_vert[3*j],s_vert[3*j+1],s_vert[3*j+2]);
                if (no_comment == 0) fprintf(fpoutv,"# Polygon (%d) around %c%c%c%c at %8.5f %8.5f %8.5f \n"
                  ,numb_list,atom_l[i][0],atom_l[i][1],atom_l[i][2],atom_l[i][3]
                  ,o_vert[3*j],o_vert[3*j+1],o_vert[3*j+2]);
                fprintf(fpoutv," Separator {\n        ");
                fprintf(fpoutv,"  Material {\n");
                fprintf(fpoutv,"   diffuseColor %s \n  }\n",plane_col_v[planeno]);
                Plane_Count++;                 /* increment number output */
                poly_list[0]=numb_list;
/* Make vertices consecutive - find center of polygon */
                xm=ym=zm=0.;
                if ((no = (int *) malloc((unsigned) ((4*numb_list) * sizeof (int)))) == NULL) {
                  printf("Unable to get no allocation\n");
                  return;
                }
                if ((ns = (int *) malloc((unsigned) ((4*numb_list) * sizeof (int)))) == NULL) {
                  printf("Unable to get ns allocation\n");
                  free(no);
                  return;
                }
                for (ii=0;ii<numb_list;ii++){
                  no[ii]=vertex_list[ii];
                  xm=xm+s_vert[3*no[ii]];
                  ym=ym+s_vert[3*no[ii]+1];
                  zm=zm+s_vert[3*no[ii]+2];
                }

                xm=xm/ (float) numb_list;
                ym=ym/ (float) numb_list;
                zm=zm/ (float) numb_list;

/* sort vertices in consecutive order based on minimal angles between vectors */

                ns[0]=no[0];
                for (ii=0;ii<numb_list-1;ii++) {    
                  phi0=1000.;
                  vxs= s_vert[3*ns[ii]]-xm;
                  vys= s_vert[3*ns[ii]+1]-ym;
                  vzs= s_vert[3*ns[ii]+2]-zm;
                  bms= sqrt( vxs*vxs+vys*vys+vzs*vzs);

                  for (jj=0;jj<numb_list;jj++) { 

                    for(ll=0;ll<ii;ll++) 
                      if (ns[ll]== no[jj]) jj++;

                    if (ns[ii]!=no[jj] && jj <numb_list){

                      vx= s_vert[3*no[jj]]-xm;
                      vy= s_vert[3*no[jj]+1]-ym;
                      vz= s_vert[3*no[jj]+2]-zm;
                      bm= sqrt( vx*vx+vy*vy+vz*vz);

                      cosphi=( vxs* vx + vys*vy+vzs*vz) /(bms*bm);
                      if (cosphi < -1. ) cosphi = -1.;
                      if (cosphi > 1. ) cosphi=1.;

                      phi = acos(cosphi)*180/3.1415926;

                      if (phi <phi0){
                        phi0=phi;
                        ns[ii+1]=no[jj];
                      }
                    }
                  } 
                } 

                draw_list=1;
                poly_list[0]=numb_list;
		for (l=0;l<numb_list;++l)
		  poly_list[draw_list++] = ns[l]+1;
		poly_list[draw_list++] = 0;       /* terminate polygon list */
                plot_vrml_poly();
                fprintf(fpoutp," /* Polygon (%d) around %c%c%c%c at %8.5f %8.5f %8.5f */ \n"
                  ,numb_list,atom_l[i][0],atom_l[i][1],atom_l[i][2],atom_l[i][3]
                  ,o_vert[3*j],o_vert[3*j+1],o_vert[3*j+2]);
                for (l=0;l<numb_list-2;++l){
                  fprintf(fpoutp," triangle {< %8.5f, %8.5f, %8.5f>,\n"
                    ,s_vert[3*ns[0]],s_vert[3*ns[0]+1],s_vert[3*ns[0]+2]);
                  fprintf(fpoutp,"<%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f>\n",
                    s_vert[3*ns[l+1]],s_vert[3*ns[l+1]+1],s_vert[3*ns[l+1]+2],
                    s_vert[3*ns[l+2]],s_vert[3*ns[l+2]+1],s_vert[3*ns[l+2]+2]);
                  fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",plane_col_p[planeno]);
                }
                if (edges == 1)  {
                  for (l=0;l<numb_list-1;++l){
                    fprintf(fpoutp," cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                      s_vert[3*ns[l]],s_vert[3*ns[l]+1],s_vert[3*ns[l]+2],
                      s_vert[3*ns[l+1]],s_vert[3*ns[l+1]+1],s_vert[3*ns[l+1]+2],radius);
                    fprintf(fpoutp,"  texture{pigment{color %s  }}\n }\n",col_edge_p);
                  } 
                  fprintf(fpoutp," cylinder { <%8.5f, %8.5f, %8.5f>, <%8.5f, %8.5f, %8.5f> , %8.5f \n",
                    s_vert[3*ns[0]],s_vert[3*ns[0]+1],s_vert[3*ns[0]+2],
                    s_vert[3*ns[numb_list-1]],s_vert[3*ns[numb_list-1]+1],s_vert[3*ns[numb_list-1]+2],radius);
                  fprintf(fpoutp," texture{pigment{color %s  }}\n }\n",col_edge_p);
                }
		draw_list = 0;
                free(no);
                free(ns);
	      }
	   }
	 }
     }
  } /* end of loop through cations */
  printf("%4d planes.\n",Plane_Count);
  fprintf(flout,"Generated %4d planes.\n",Plane_Count);
}

/* ************************************************************** */
/* ************************************************************** */

void generate_poly(output)

GEOMedit_list *output;

/* procedure to generate vertices of polyhedra */

{
  extern int natom;
  extern int npoly;           /* number of different types of polyhedra */
  int i,j,k;                  /* loop counters, etc */
  int nvert_start;            /* starting number for cations */
  int polyno;                 /* number of current polyhedron */
  int Poly_Count;             /* Number of polyhedra output */
  int Max_Vertices;           /* Maximum number of vertices */
  float radius;

  radius = rad_edge * 0.01 * Scale;
  if (npoly == 1) return;     /* exit NOW if no work */
  Poly_Count = 0;
  for (i=0;i<3;++i) boxlim[i] += 3.0; /* expand box for anion generation */
  nvert = 0;                  /* make list empty */
  for (i=0;i<natom;++i)       /* loop through atoms */
  if (atom_n[i] < 1) find_all_in_box(i); /* generate all anions */
  for (i=0;i<3;++i) boxlim[i] -= 3.0; /*shrink box to original size */
  if (nvert == 0) return;     /* if no anions, no polyhedra */
  nvert_start = nvert;        /* save start of cation list */
  for (i=0;i<natom;++i) {     /* start loop for cations */
    nvert = nvert_start;      /* get rid of previous cations */
    if (atom_n[i] > 0 && atom_n[i] <10000) {    /* greater than zero for polyhedral centers */
      polyno = atom_n[i];     /* save polyhedron number */
      find_all_in_box(i);   /* add all polyhedral centers in box */
      if(nvert_start < nvert) { /* check for some found */
        draw_list = 0;        /* initialize the draw list */
        Max_Vertices = 0;
        for (j=nvert_start;j<nvert;++j) { /* first loop through cations */
	  numb_list = 0;      /* list clear */
	  for (k=0;k<nvert_start;++k) { /* loop through anions */
	    if (dist(j,k) <= poly_size[polyno])  /* add vertex k to poly list*/
	     numb_list++;
	     /* vertex_list[numb_list++] = k;*/
	  }
          if ( numb_list > Max_Vertices) Max_Vertices = numb_list;
        }   /* end of first pass through list */
	draw_list = 0;  /* initialize the draw list */
        for (j=nvert_start;j<nvert;++j) { /* second loop through cations */
	  numb_list = 0;      /* list clear */
	  for (k=0;k<nvert_start;++k) { /* loop through anions */
	    if (dist(j,k) <= poly_size[polyno])  /* add vertex k to poly list*/
	      vertex_list[numb_list++] = k;
	  }
          if (numb_list >3 && numb_list == Max_Vertices) { /* only do if polyhedron complete */
            fprintf(flout,"Polyhedron with %d vertices around %c%c%c%c at %8.5f %8.5f %8.5f \n"
              ,numb_list,atom_l[i][0],atom_l[i][1],atom_l[i][2],atom_l[i][3]
              ,s_vert[3*j],s_vert[3*j+1],s_vert[3*j+2]);
            fprintf(fpoutp," /* Polyhedron (%d) around %c%c%c%c at %8.5f %8.5f %8.5f */ \n"
              ,numb_list,atom_l[i][0],atom_l[i][1],atom_l[i][2],atom_l[i][3]
              ,o_vert[3*j],o_vert[3*j+1],o_vert[3*j+2]);
            if (no_comment == 0) fprintf(fpoutv,"# Polyhedron (%d) around %c%c%c%c at %8.5f %8.5f %8.5f \n"
              ,numb_list,atom_l[i][0],atom_l[i][1],atom_l[i][2],atom_l[i][3]
              ,o_vert[3*j],o_vert[3*j+1],o_vert[3*j+2]);
            fprintf(fpoutv," Separator {\n    ");
            fprintf(fpoutv,"  Material {\n   diffuseColor  %s \n  }\n",poly_col_v[polyno]);
	    build_poly_list(numb_list,polyno); /* select polygons */
	    poly_list[draw_list++] = 0;       /* terminate polygon list */
            plot_vrml_poly(); /* pov part already done in build_poly_list */
	    draw_list = 0;
            Poly_Count++;
	  } 
	}
      }
    }
  } /* end of loop through cations */
  printf("%4d polyhedra.\n",Poly_Count);
  fprintf(flout,"Generated %4d polyhedra.\n",Poly_Count);
}

/* ************************************************************** */
/* ************************************************************** */

void import_gsas(input,in_line)

char input[100];
int in_line;

/* routine to extract structural information from a GSAS exp file */

{
  char string[100],filename[100],temp[5],tstring[20];
  int phase_no,i;
  FILE *impin;

  if (in_line) {
  fprintf(stderr,"Cannot read inline GSAS data. Please use the import instruction\n");
#ifdef _MAC_
  printf("Use Apple Q to close this window.\n");
#endif
  exit(0);
  }
  sscanf(input,"%s %s %d",filename,filename,&phase_no); /* get file name and phase number */
  if ((phase_no < 1) && (phase_no > 9)) {
    fprintf(stderr,"Illegal phase number in GSAS import instruction, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
    exit(0);
  }
  if ((impin=fopen(filename,"r")) == NULL) {
    fprintf(stderr,"Cannot Open GSAS Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
    exit(0);
  }
  sprintf(temp,"CRS%d",phase_no);  /* make search object */
  while(strncmp(string,temp,4)!=0)if(!fgets(string,100,impin)){; /* search for phase info */
    fprintf(stderr,"Error reading GSAS Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
    exit(0);
  }
  for(;;) { /* search for lattice constants */
    if(!fgets(string,100,impin)){ 
      fprintf(stderr,"Error reading GSAS Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    sscanf(string,"%s %s",tstring,tstring);
    if(strncmp(tstring,"ABC",3)==0) {
      sscanf(string,"%s %s %f %f %f",tstring,tstring,&lat_con[0],&lat_con[1]
         ,&lat_con[2]);     /* read lattice constants */
      break;
    }
  }
  for(;;) {  /* search for angles */
    if(!fgets(string,100,impin)){
      fprintf(stderr,"Error reading GSAS Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    sscanf(string,"%s %s",tstring,tstring);
    if(strncmp(tstring,"ANGLES",6)==0) {
      sscanf(string,"%s %s %f %f %f",tstring,tstring,&lat_con[3],&lat_con[4]
         ,&lat_con[5]);     /* read lattice constants */
      break;
    }
  }
  fprintf(flout,"******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n"
     ,lat_con[0],lat_con[1],lat_con[2],lat_con[3],lat_con[4],lat_con[5]);
  for(;;) {   /* position to atomic coordinates */
    if(!fgets(string,100,impin)){
      fprintf(stderr,"Error reading GSAS Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    sscanf(string,"%s %s",tstring,tstring);
    if(strncmp(tstring,"AT",2)==0) break;
  }
/* at this point, positioned at first atom line */
  for(;;) {  /* process atom parameters */
    sscanf(string,"%s %s",tstring,tstring);
    if(strncmp(tstring,"AT",2) !=0 ) break;  /* out when atom stuff done */
    sscanf(string,"%s %s %s %s %f %f %f",tstring,tstring,tstring,tstring,
      &atom_xyz[natom][0],&atom_xyz[natom][1],&atom_xyz[natom][2]);
    atom_n[natom] = 1;
    for (i=0;i<4;i++) atom_l[natom][i] = ' ';  /* initialize atom name */
    if (string[62] >= '0' && string[62] <= '9')
      atom_n[natom] = (int)string[62]-48;
    else
      atom_l[natom][0] = string[62];
    if (string[63] >= '0' && string[63] <= '9')
      atom_n[natom] = 10 * atom_n[natom] + (int)string[63]-48;
    else
      atom_l[natom][1] = string[63];
    if (string[64] >= '0' && string[64] <= '9')
      atom_n[natom] = 10 * atom_n[natom] + (int)string[64]-48;
    else
      atom_l[natom][2] = string[64];
    if (string[65] >= '0' && string[65] <= '9')
      atom_n[natom] = 10 * atom_n[natom] + (int)string[65]-48;
    else
      atom_l[natom][3] = string[65];
    fprintf(flout,"******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n",atom_l[natom][0]
      ,atom_l[natom][1],atom_l[natom][2],atom_l[natom][3],atom_n[natom]
      ,atom_xyz[natom][0],atom_xyz[natom][1],atom_xyz[natom][2]);
    fgets(string,100,impin);
    if (string[74] == 'A' ) {
      ell_type[n_ellips] = 1; /* Uij */
      do_ellipsoids = 1;
      sscanf(string,"%s %s %s %f %f %f %f %f %f",tstring,tstring,tstring,
        &ellips[n_ellips][0],&ellips[n_ellips][1],&ellips[n_ellips][2],
        &ellips[n_ellips][3],&ellips[n_ellips][4],&ellips[n_ellips][5]);  /* read unique coefficients */
      for(i=0;i<4;i++)
         ellips_l[n_ellips][i] = atom_l[natom][i];
      ellips_n[n_ellips] = atom_n[natom];
      fprintf(flout,"******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n"
        ,atom_l[natom][0],atom_l[natom][1],atom_l[natom][2],atom_l[natom][3]
        ,atom_n[natom],ellips[n_ellips][0],ellips[n_ellips][1],ellips[n_ellips][2]
        ,ellips[n_ellips][3],ellips[n_ellips][4],ellips[n_ellips][5]);
      if (++n_ellips >= MAX_EL) {
        fprintf(flout,"Too many ellipsoid types specified - increase MAX_EL.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
        exit(0);
      }
    }
    if (++natom >= MAX_ATOM) {
      fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
    if(!fgets(string,100,impin)){
      fprintf(stderr,"Error reading GSAS Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
  }
  for (;;) {   /* start loop for Space Group information */
    if(!fgets(string,100,impin)){
      fprintf(stderr,"Error reading GSAS Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);                          
    }
    sscanf(string,"%s %s",tstring,tstring);
    if(strncmp(tstring,"SG",2)==0) break;  /* out when space group found */
  }
  string[0] = 's';
  string[1] = 'p';
  string[2] = 'g';
  string[3] = 'r';
  for (i=3;i<60;i++)
    string[i+1] = string[i+10];
  string[60] = '\0';
  fprintf(flout,"******      %s\n",string);
  symop(string);
  fclose(impin);
}

/* ************************************************************** */
/* ************************************************************** */

void import_shelx(input,in_line)

char input[100];
int in_line; /*flag for in-line shelx data */

/* routine to extract structural information from a SHELX data file */

{
  FILE *impin;
  char string[100];              /* temporary string */
  char filename[100];            /* string for filename */
/*  char symmop[50];        */       /* temporary string for SYMM */
/*  char free_params[50]; */         /* temporary string for FVAR */
  int intype;                    /* type of datum read */
  int i,j;                       /* temporary */
  float xray;                    /* xray wavelength (read, but not used) */                   
  int read_start;                /* flag for start of atom list */  
  int read_uij;                  /* flag for uij card */  
  char marker[2];                /* continuation marker on atom card */
  float u11,u22;                 /* temporary storage for uij */
  float site_occup;              /* occupancy (read, but not used) */ 
  static char shcmds[][67]={"ZERR","LATT","SYMM","SFAC","DISP","UNIT","LAUE",
                         "REM ","MORE","TIME","OMIT","SHEL","BASF","TWIN",
                         "EXTI","SWAT","HOPE","MERG","SPEC","RESI","MOVE",
                         "ANIS","AFIX","HFIX","FRAG","FEND","EXYZ","EADP",
                         "EQIV","CONN","PART","BIND","FREE","DFIX","DANG",
                         "BUMP","SAME","SADI","CHIV","FLAT","DELU","SIMU",
                         "DEFS","ISOR","NCSY","SUMP","L.S.","CGLS","BLOC",
                         "DAMP","STIR","WGHT","FVAR","BOND","CONF","MPLA",
                         "RTAB","HTAB","LIST","ACTA","SIZE","TEMP","WPDB",
                         "FMAP","GRID","PLAN","MOLE"};
  static char cmd[]="    ";            /*temporary storage for start of line*/
  
  if (in_line) {
  impin=fpin;
  } else {
  sscanf(input,"%s %s",filename,filename); /* get file name */
  if ((impin=fopen(filename,"r")) == NULL) {
    fprintf(stderr,"Cannot Open SHELX Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
    exit(0);
  }
  }
  read_start = 0;
  read_uij = 0;
  fprintf(flout,"****** Now interpreting SHELX-style data\n");
  while (!feof(impin)) {
    if(!fgets(input,100,impin)) {             /* read a line */
      fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(1);
    }
    intype = 0;
    for(i=0;i<4;i++) {
      cmd[i]=(char)toupper(input[i]);
      if(input[i] == '\0')    intype = 3;
    }

    if(strcmp(cmd,"TITL") == 0)intype = 1; /* echo and ignore */
    if(strcmp(cmd,"CELL") == 0)intype = 2;

    if(strcmp(cmd,"HKLF") == 0)intype = 4; /* trigger exit */
    if(strcmp(cmd,"END ") == 0)intype = 4; /* trigger exit*/
 
    if (intype ==0 ) { /* check against list of 'other' commands */
    for (i=0;i<67;i++) {
              if (strcmp(cmd,shcmds[i]) ==0) {
              intype=3;
              break;
              }
        }
    }
    switch (intype) {
        case 1:
            for(i=0;i<51;i++) if(input[i] == '\n')input[i] = ' ';
            input[50] = '\0';    /* truncate title */
            fprintf(flout,"******      %s\n",input);
            break;      /* ignore title except for listing */
        case 2: for (i=3;i<=5;++i) lat_con[i] = 0.;
            sscanf(input,"%s %f %f %f %f %f %f %f",string,&xray,&lat_con[0]
	      ,&lat_con[1],&lat_con[2],&lat_con[3],&lat_con[4]
 	      ,&lat_con[5]);
            fprintf(flout,"******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n"
              ,lat_con[0],lat_con[1],lat_con[2],lat_con[3],lat_con[4],lat_con[5]);
	    break;
        case 3:    /* just ignore most SHELX commands */
            break;    
        case 4:
            read_start = 0;
            fprintf(flout,"****** HKLF or END - end of SHELX dataset\n");
            if (!in_line) fclose(impin);
            return;
            break;
        default:           
            sscanf(input,"%c%c%c%c %d %f %f %f %f %f %f %s",
              &atom_l[natom][0],&atom_l[natom][1]
             ,&atom_l[natom][2],&atom_l[natom][3],
              &atom_n[natom],&atom_xyz[natom][0],
              &atom_xyz[natom][1],&atom_xyz[natom][2],
              &site_occup,&u11,&u22,marker);
            for (j=0;j<3;j++)
              if (atom_xyz[natom][j] > 9. ) atom_xyz[natom][j] -= 10.;
            atom_n[natom] = 0;
            for (j=0;j<4;j++)   /* extract atom number if part of SHELX name */
              if (atom_l[natom][j] >= '0' && atom_l[natom][j] <= '9') {
              atom_n[natom] = 10* atom_n[natom]+(int)atom_l[natom][j]-48;
                atom_l[natom][j] = ' ';
              }
              if (atom_n[natom] == 0) atom_n[natom]=1;
              
            fprintf(flout,"******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n"
              ,atom_l[natom][0],atom_l[natom][1],atom_l[natom][2],atom_l[natom][3]
              ,atom_n[natom],atom_xyz[natom][0],atom_xyz[natom][1],atom_xyz[natom][2]);
            if (strncmp(marker,"=",1) == 0) {
                  strcpy(marker," ");
              if(!fgets(input,100,impin)) {             /* read a line */
                fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
                exit(1);
              }
              sscanf(input," %f %f %f %f",
                &ellips[n_ellips][2],&ellips[n_ellips][5],
                &ellips[n_ellips][4],&ellips[n_ellips][3]);
              ell_type[n_ellips] = 1; /* Uij */
              for (i=0;i<4;i++)
                ellips_l[n_ellips][i] = atom_l[natom][i];
              ellips_n[n_ellips] = atom_n[natom];
              ellips[n_ellips][0]=u11;
              ellips[n_ellips][1]=u22;
              fprintf(flout,"******      Uij  %c%c%c%c%3d %8.5f %8.5f %8.5f %8.5f %8.5f %8.5f\n"
                ,atom_l[natom][0],atom_l[natom][1],atom_l[natom][2],atom_l[natom][3]
                ,atom_n[natom],ellips[n_ellips][0],ellips[n_ellips][1],ellips[n_ellips][2]
                ,ellips[n_ellips][3],ellips[n_ellips][4],ellips[n_ellips][5]);
              n_ellips++;
              if (n_ellips >= MAX_EL) {
                fprintf(flout,"Too many ellipsoid types specified - increase MAX_EL.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
                exit(0);
              }
            }
            natom++;
            if (natom >= MAX_ATOM) {
              fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
              exit(0);
            }
            break;    /* end of default */
    }   /* end of switch */
  }
if (!in_line)  fclose(impin);
}  /* end of import_shelx */

/* ************************************************************** */
/* ************************************************************** */

void import_fdat(input,in_line)

char input[100];
int in_line; /*flag for in-line data */

/* routine to extract structural information from a CSD 'FDAT' data file */

{
  FILE *impin;
  char string[100];       /* temporary string */
  char filename[100];     /* string for filename */
  int i,j,k;              /* temporary */
  char spstring[20];      /* space group symbol */
  char lat_s[6][6];       /*temporary storage for lattice constants*/ 
  int ldec[6];            /*temporary storage for decimal places of lattice constants*/ 
  char number[4];
  
  memset(spstring,0,20);
  memset(string,0,100);
  if (in_line) {
    impin=fpin;
  } else {
    sscanf(input,"%s %s",filename,filename); /* get file name */
    if ((impin=fopen(filename,"r")) == NULL) {
      fprintf(stderr,"Cannot Open CSD FDAT Import File, Run aborted.\n");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(0);
    }
  }
  
  fprintf(flout,"****** Now interpreting CSD-style data\n");

  if(!fgets(input,100,impin)) {             /* read a line */
    fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
    exit(1);
  }

/* the first line should contain the REFCODE of the database entry */
/* but we need not rely on this as this line does not contain any  */
/* other useful information */

  if(strncmp(input,"#",1) == 0) {
    input[0]= ' ';
    input[9] = '\0';    /* truncate title */
    fprintf(flout,"******      %s\n",input);

    if(!fgets(input,100,impin)) {             /* read a line */
      fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
      exit(1);
    }
  }
   
/* lattice constants come in a fixed FORTRAN format followed by lots */
/* of trailing integers and the spacegroup symbol */

  for (i=0;i<=5;++i) lat_con[i] = 0.;

#ifdef _PC_
  sscanf(input,"%6c%6c%6c%6c%6c%6c%1d%1d%1d%1d%1d%1d%21c%s",lat_s[0],
    lat_s[1],lat_s[2],lat_s[3],lat_s[4],lat_s[5],&ldec[0],&ldec[1],&ldec[2],
    &ldec[3],&ldec[4],&ldec[5],string,string);
#else
  sscanf(input,"%6c%6c%6c%6c%6c%6c%1d%1d%1d%1d%1d%1d%*[0-9 ]%s",lat_s[0],
    lat_s[1],lat_s[2],lat_s[3],lat_s[4],lat_s[5],&ldec[0],&ldec[1],&ldec[2],
    &ldec[3],&ldec[4],&ldec[5],string);
#endif

  for (i=0;i<6;++i) {
    lat_con[i]=atof(lat_s[i])/pow(10,ldec[i]);
  }
          
/* separate class symbol from other symmetry elements */   
  spstring[0] = 's';
  spstring[1] = 'p';
  spstring[2] = 'g';
  spstring[3] = 'r';
  spstring[4] = ' ';
  spstring[5]=string[0];
  j=5;
  k=0;
/* explicit treatment of those spacegroups that the formalism below cannot
handle*/
/* yes - this IS bad ... FIX ME*/
  if (!strcmp(string,"P42")) { strcpy(spstring,"spgr P 42");
  } else 
    if (!strcmp(string,"P41")) { strcpy(spstring,"spgr P 41");
    } else 
    if (!strcmp(string,"P43")) { strcpy(spstring,"spgr P 43");
    } else 
      if (!strcmp(string,"P42/m")) { strcpy(spstring,"spgr P 42/m");
    } else 
    if (!strcmp(string,"P42/n")) { strcpy(spstring,"spgr P 42/n");
    } else 
    if (!strcmp(string,"P4222")) { strcpy(spstring,"spgr P 42 2 2");
    } else 
    if (!strcmp(string,"P42212")) { strcpy(spstring,"spgr P 42 21 2");
    } else 
    if (!strcmp(string,"P4122")) { strcpy(spstring,"spgr P 41 2 2");
    } else 
    if (!strcmp(string,"P4322")) { strcpy(spstring,"spgr P 43 2 2");
    } else 
    if (!strcmp(string,"P41212")) { strcpy(spstring,"spgr P 41 21 2");
    } else 
    if (!strcmp(string,"P43212")) { strcpy(spstring,"spgr P 43 21 2");
    } else 
    if (!strcmp(string,"P42cm")) { strcpy(spstring,"spgr P 42 c m");
    } else 
    if (!strcmp(string,"P42nm")) { strcpy(spstring,"spgr P 42 n m");
    } else 
    if (!strcmp(string,"P42mc")) { strcpy(spstring,"spgr P 42 m c");
    } else 
    if (!strcmp(string,"P42bc")) { strcpy(spstring,"spgr P 42 b c");
    } else 
    if (!strcmp(string,"P42/mmc")) { strcpy(spstring,"spgr P 42/m m c");
    } else 
    if (!strcmp(string,"P42/mcm")) { strcpy(spstring,"spgr P 42/m c m");
    } else 
    if (!strcmp(string,"P42/nbc")) { strcpy(spstring,"spgr P 42/n b c");
    } else 
    if (!strcmp(string,"P42/nnm")) { strcpy(spstring,"spgr P 42/n n m");
    } else 
    if (!strcmp(string,"P42/mnm")) { strcpy(spstring,"spgr P 42/m n m");
    } else 
    if (!strcmp(string,"P42/nmc")) { strcpy(spstring,"spgr P 42/n m c");
    } else 
    if (!strcmp(string,"P42/ncm")) { strcpy(spstring,"spgr P 42/n c m");
    } else 
    if (!strcmp(string,"P32")) { strcpy(spstring,"spgr P 32");
    } else 
    if (!strcmp(string,"P63")) { strcpy(spstring,"spgr P 63");
    } else 
    if (!strcmp(string,"P62")) { strcpy(spstring,"spgr P 62");
    } else 
    if (!strcmp(string,"P64")) { strcpy(spstring,"spgr P 64");
    } else 
    if (!strcmp(string,"P63/m")) { strcpy(spstring,"spgr P 63/m");
    } else 
    if (!strcmp(string,"P6222")) { strcpy(spstring,"spgr P 62 2 2");
    } else 
    if (!strcmp(string,"P6422")) { strcpy(spstring,"spgr P 64 2 2");
    } else 
    if (!strcmp(string,"P6322")) { strcpy(spstring,"spgr P 63 2 2");
    } else 
    if (!strcmp(string,"P63cm")) { strcpy(spstring,"spgr P 63 c m");
    } else 
    if (!strcmp(string,"P63mc")) { strcpy(spstring,"spgr P 63 m c");
    } else 
    if (!strcmp(string,"P63/mcm")) { strcpy(spstring,"spgr P 63/m c m");
    } else 
    if (!strcmp(string,"P63/mmc")) { strcpy(spstring,"spgr P 63/m m c");
    } else 
    if (!strcmp(string,"P4232")) { strcpy(spstring,"spgr P 42 3 2");
    } else 
    if (!strcmp(string,"P4132")) { strcpy(spstring,"spgr P 41 3 2");
    } else 
    if (!strcmp(string,"P4332")) { strcpy(spstring,"spgr P 43 3 2");
    } else 
    { 
      for (i=1;i<=strlen(string);i++) {
        switch (string[i]) {
        case ' ':         
          break;
        case '1': 
        case '5':
        case '4':        
          j++;
          spstring[j]=string[i];
          break;
        case '/':          
/*        j++;
          spstring[j]=' '; */
          j++;
          spstring[j]='/';
          k=1;
          break;
        default:
          j++;
          if (k==0) {
            spstring[j]=' ';
            j++;
          }
          spstring[j]=string[i];
          k=0;
          break;
        }   /*switch*/        
      } /*for i*/
      spstring[j+1]='\0';
    }
           
    fprintf(flout,"******      Space group *%s*\n",spstring);
    symop(spstring); 
        
    fprintf(flout,"******      Cell     %8.5f %8.5f %8.5f %8.3f %8.3f %8.3f\n"
        ,lat_con[0],lat_con[1],lat_con[2],lat_con[3],lat_con[4],lat_con[5]);

/*skip R value and first bit screen*/
    for (i=0;i<2;i++){
      if(!fgets(input,100,impin)) {             /* read a line */
        fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
        exit(1);
      }
    }

/* skip CSD bit screens and sum formula */ 
      while (input[0] >= '0' && input[0] <= '9') {
/*    while (isdigit(input[0])){ */
      if(!fgets(input,100,impin)) {             /* read a line */
        fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
        exit(1);
      }
    }
        
/* read coordinates until EOF or empty atom symbol encountered */
    while (!feof(impin)) {
     
      if(!fgets(input,100,impin)) {             /* read a line */
        fprintf(flout,"Error reading file.");
#ifdef _MAC_
      printf("Use Apple Q to close this window.\n");
#endif
        exit(1);
      }

/* atoms come in groups of three */
      sscanf(input,"%c%c%c%c %7f %7f %7f %c%c%c%c %7f %7f %7f %c%c%c%c %7f %7f %7f",
              &atom_l[natom][0],&atom_l[natom][1],
              &atom_l[natom][2],&atom_l[natom][3],
              &atom_xyz[natom][0],&atom_xyz[natom][1],&atom_xyz[natom][2],
              &atom_l[natom+1][0],&atom_l[natom+1][1],
              &atom_l[natom+1][2],&atom_l[natom+1][3],
              &atom_xyz[natom+1][0],&atom_xyz[natom+1][1],&atom_xyz[natom+1][2],
              &atom_l[natom+2][0],&atom_l[natom+2][1],
              &atom_l[natom+2][2],&atom_l[natom+2][3],
              &atom_xyz[natom+2][0],&atom_xyz[natom+2][1],&atom_xyz[natom+2][2]);
              
/* and the atom list is terminated by numerical data */      
       if (atom_l[natom][0] < 'A') {
         natom--;
         if (!in_line)  fclose(impin);
         return;
       }
                
/* scale coordinates */
       for (j=0;j<3;j++) {
         atom_xyz[natom][j] = atom_xyz[natom][j]/100000.;
         atom_xyz[natom+1][j] = atom_xyz[natom+1][j]/100000.;
         atom_xyz[natom+2][j] = atom_xyz[natom+2][j]/100000.;
       }

/* separate numbers from labels and remove trailing identifiers */

       for (i=0;i<3;i++){
         k=0;
         for (j=1;j<4;j++) {   
           number[k]=' ';
             if (atom_l[natom+i][j] >= '0' && atom_l[natom+i][j] <= '9') {
/*           if ( isdigit(atom_l[natom+i][j]) ) { */
             number[k]=atom_l[natom+i][j];
             k++;
           }
           if (k) atom_l[natom+i][j] = ' ';
         }
         atom_n[natom+i]=atoi(number);
       }
 
       if (Unique_Atom() == 1) {
/* report converted data of first atom */
         fprintf(flout,"******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n"
           ,atom_l[natom][0],atom_l[natom][1],atom_l[natom][2],atom_l[natom][3]
           ,atom_n[natom],atom_xyz[natom][0],atom_xyz[natom][1],atom_xyz[natom][2]);
         natom++;
         if (natom >= MAX_ATOM) {   /* first unique */
           fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
           printf("Use Apple Q to close this window.\n");
#endif
           exit(0);
         }
       } /* first unique */
/* return now if no second atom on the line */    
       if (atom_l[natom][0] == '\0' ) return;
       if (Unique_Atom() == 1) {
/* report converted data of second atom */    
         fprintf(flout,"******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n"
           ,atom_l[natom][0],atom_l[natom][1],atom_l[natom][2],atom_l[natom][3]
           ,atom_n[natom],atom_xyz[natom][0],atom_xyz[natom][1],atom_xyz[natom][2]);
         natom++;
         if (natom >= MAX_ATOM) {    /* second unique */
           fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
           printf("Use Apple Q to close this window.\n");
#endif
           exit(0);
         }
       } /* second unique */
/* return now if no third atom on the line */    
       if (atom_l[natom][0] == '\0') return;
       if (Unique_Atom() == 1) {     /* third unique */
/* report converted data of third atom */    
         fprintf(flout,"******      Atom %c%c%c%c%3d %8.5f %8.5f %8.5f\n"
           ,atom_l[natom][0],atom_l[natom][1],atom_l[natom][2],atom_l[natom][3]
           ,atom_n[natom],atom_xyz[natom][0],atom_xyz[natom][1],atom_xyz[natom][2]);
         natom++;
         if (natom >= MAX_ATOM) {
           fprintf(flout,"Too many atoms in asymmetric unit - increase MAX_ATOM.\n"); 
#ifdef _MAC_
           printf("Use Apple Q to close this window.\n");
#endif
           exit(0);
         }
       }  /* third unique */
     } /* while !feof */

     if (!in_line)  fclose(impin);

  }  /* end of import_fdat */

/* ************************************************************** */
/* ************************************************************** */

int Unique_Atom()

/* routine to check if atom at position 'natom' is unique. Returns 1 if it
   is. If the position is a duplicate, returns 0. */

{
  int j,k,test;

  for (k=0;k<natom-1;k++) { /* loop through atoms */
    if (atom_n[k] == atom_n[natom]){
      test = 0;
      for (j=0;j<4;j++) {
         if (atom_l[k][j] != atom_l[natom][j]) test = 1;
      }
      if (test == 0) {      /* name and number the same - check coordinates */
        expand_atom(natom); /* find all of test atoms in unit cell */
        for (j=0;j<ncell;j++) {
          if ((fabs(atom_xyz[k][1] - cell_xyz[j][1]) < 0.0001) &&
              (fabs(atom_xyz[k][2] - cell_xyz[j][2]) < 0.0001) &&
              (fabs(atom_xyz[k][0] - cell_xyz[j][0]) < 0.0001)) return (0); /* duplicate */
        }
      }
    }
  }
  return(1);    /* unique */
}
